//
//  PipelineStore.swift
//  Spokestack Studio iOS
//
//  Created by Daniel Tyreus on 4/13/20.
//  Copyright © 2020 Spokestack. All rights reserved.
//

import SwiftUI
import Spokestack

class PipelineStore: ObservableObject {
    
    @Published var isListening: Bool = false;
    @Published var isSpeaking: Bool = false;
    @Published var userSays: String? = nil;
    @Published var appSays: String? = nil;

    private var dialogManager:SkillDialogManager;
    private var nluService = NLUService("nlu")
    
    private var currentResponse:Response = Response(speak:"")
    
    lazy private var tts: TextToSpeech = {
        return TextToSpeech(self, configuration: SpeechConfiguration())
    }()
    
    
    init(_ dialogManager:SkillDialogManager) {
        self.dialogManager = dialogManager
    }
    
    enum PipelineMode {
        case push2talk
        case wakeword
    }
    
    private var mode:PipelineMode = .push2talk
    
    lazy private var pipeline: SpeechPipeline = {
       
        return SpeechPipeline(self, pipelineDelegate: self)
    }()
    
    func startPipeline() {
        print("[\(mode)] starting pipeline")
        self.pipeline.start()
    }

    func stopPipeline() {
        print("[\(mode)] stopping pipeline")
        self.pipeline.stop()
    }
    
    func activatePipeline() {
        print("[\(mode)] manually activating pipeline")
        self.pipeline.activate()
        DispatchQueue.main.async {
            self.isListening = true
        }
        
    }
    
    func deactivatePipeline() {
        print("[\(mode)] manually deactivating pipeline")
        self.pipeline.deactivate()
        DispatchQueue.main.async {
            self.isListening = false
        }
    }
    
    func speak(_ request:Response) {
        DispatchQueue.main.async {
            self.appSays = request.speak
            self.isSpeaking = true
        }
        self.currentResponse = request
        self.tts.speak(TextToSpeechInput(request.speak))
    }
    
}

extension PipelineStore: SpeechEventListener {
    func didActivate() {
        print("[\(mode)] heard wakeword")
        if (mode == .wakeword) {
            self.pipeline.activate()
            DispatchQueue.main.async {
                self.isListening = true
            }
        }
    }
    
    func didDeactivate() {
        print("[\(mode)] heard speech ended")
        //this gets called after the user stops talking
        //we need to actually deactive the pipeline next
        self.pipeline.deactivate()
        DispatchQueue.main.async {
            self.isListening = false
        }
    }
    
    func didError(_ error: Error) {
        print("[\(mode)] did didError \(error)")
        currentResponse = try! self.dialogManager.turn(type: "UnknownRequest", error: error)
        self.appSays = currentResponse.speak
        self.isSpeaking = true
        self.tts.speak(TextToSpeechInput(currentResponse.speak))
    }
    
    func didTrace(_ trace: String) {
        print("[\(mode)] did trace \(trace)")
    }
    
    func didRecognize(_ result: SpeechContext) {
        
        print("[\(mode)] didRecognize \(result.isSpeech) and transscript \(result.transcript)")
        DispatchQueue.main.async {
            self.appSays = nil;
            self.userSays = result.transcript
        }
        
        let _ = nluService.nlu!.classify(utterances: [result.transcript])
        .subscribe(on: DispatchQueue.global(qos: .userInitiated))
        .sink(receiveCompletion: { completion in
            switch completion {
            case .failure(let error):
                // respond appropriately to an error in classification
                print("nlu failure \(error)")
                break
            case .finished:
                // respond appropriately to finished classification
                print("nlu finished")
                break
            }
        }, receiveValue: { results in
            let _ = results.map({ result in
                print("nlu result")
                print(result)
                self.currentResponse = try! self.dialogManager.turn(type: "IntentRequest", nluResult: result)
                print(self.currentResponse.speak)
                DispatchQueue.main.async {
                    self.appSays = self.currentResponse.speak
                    self.isSpeaking = true
                }
                self.tts.speak(TextToSpeechInput(self.currentResponse.speak))
                
            })
        })
    }

    func didTimeout() {
        print("[\(mode)] didTimeout")
        currentResponse = try! self.dialogManager.turn(type: "TimeoutRequest")
        print(currentResponse.speak)
        DispatchQueue.main.async {
            self.userSays = nil
            self.appSays = self.currentResponse.speak
            self.isSpeaking = true
        }
        self.tts.speak(TextToSpeechInput(currentResponse.speak))
    }
    
    func failure(speechError: Error) {
        print("asr failed \(speechError)")
        DispatchQueue.main.async {
            self.isListening = false
        }
    }

}

extension PipelineStore: PipelineDelegate {
    func didInit() {
        print("[\(mode)] didInit")
    }
    
    func didStart() {
        print("[\(mode)] didStart")
    }
    
    func didStop() {
        print("[\(mode)] didStop")
    }
    
    func setupFailed(_ error: String) {
        print("[\(mode)] setup failed \(error)")
    }
    
}


extension PipelineStore: TextToSpeechDelegate {
    
    func success(result: TextToSpeechResult) {
        print("tts success")
    }
    
    func failure(ttsError: Error) {
        print("tts failed \(ttsError)")
        DispatchQueue.main.async {
            self.isSpeaking = false
        }
    }
    
    func didBeginSpeaking() {
        print("didBeginSpeaking")
    }
    
    func didFinishSpeaking() {
        print("didFinishSpeaking")
        DispatchQueue.main.async {
            self.isSpeaking = false
        }
        
        if (currentResponse.reprompt != nil) {
            print("opening mic")
            activatePipeline();
        }
    }
    
    
}

struct PipelineStore_Previews: PreviewProvider {
    static var previews: some View {
        /*@START_MENU_TOKEN@*/Text("Hello, World!")/*@END_MENU_TOKEN@*/
    }
}
