//
//  SkillDialogManager.swift
//  Minecraft Skill Demo
//
//  Created by Daniel Tyreus on 5/5/20.
//  Copyright © 2020 Spokestack. All rights reserved.
//

import Fuse
import Spokestack

struct HandlerInput {
    
    var responseBuilder: ResponseBuilder
    var attributeMananger: AttributeManager
    var requestEnvelope: RequestEnvelope
    
    init(session: [String:Any], type: String, nluResult: NLUResult?) {
        responseBuilder = ResponseBuilder()
        attributeMananger = AttributeManager(session: session)
        requestEnvelope = RequestEnvelope(type: type, nluResult: nluResult)
    }
}

struct Response {
    var speak: String
    var reprompt: String?
}

protocol RequestHandler {
    func canHandle(handlerInput:HandlerInput) -> Bool
    func handle(handlerInput:HandlerInput) -> Response
}

struct RequestEnvelope {
    struct Intent {
        var name: String
        var slots: [String:Slot]?
    }
    struct Request {
        var type: String
        var intent: Intent?
    }
    
    var request: Request
    
    init( type: String, nluResult: NLUResult?) {
        request = Request(type: type)
        if (nluResult != nil) {
            request.intent = Intent(name: nluResult!.intent, slots: nluResult!.slots)
        }
    }
}

class ResponseBuilder {
    
    var response: Response
    
    init() {
        response = Response(speak:"")
    }
    
    func speak(_ speechOutput: String) -> ResponseBuilder {
        self.response.speak = speechOutput
        return self
    }
    
    @discardableResult
    func reprompt(_ repromptSpeechOutput: String) -> ResponseBuilder {
        self.response.reprompt = repromptSpeechOutput
        return self
    }
    
    func getResponse() -> Response {
        return response
    }
}

class AttributeManager {
    
    var _session:[String:Any]
    
    init(session:[String:Any]) {
        _session = session
    }

    func getSessionAttributes() -> [String:Any] {
        return _session
    }
    
    func setSessionAttributes(sessionAttributes : [String:Any]) {
        _session = _session.merging(sessionAttributes, uniquingKeysWith: { (_, last) in last })
    }
}


struct HelloWorldIntentHandler : RequestHandler {
  func canHandle(handlerInput : HandlerInput) -> Bool {
    return handlerInput.requestEnvelope.request.type == "IntentRequest"
      && handlerInput.requestEnvelope.request.intent!.name == "HelloWorldIntent";
  }
  func handle(handlerInput : HandlerInput) -> Response {
    let speechText = "Hello World!";

    return handlerInput.responseBuilder
      .speak(speechText)
      .getResponse()
  }
}



struct LaunchRequestHandler: RequestHandler {
    func canHandle(handlerInput: HandlerInput) -> Bool {
        return handlerInput.requestEnvelope.request.type == "LaunchRequest"
    }
    
    func handle(handlerInput: HandlerInput) -> Response {
        let item:String = recipes.randomElement()!.key
        let speak = String(format: responses["WELCOME_MESSAGE"]!, responses["SKILL_NAME"]! ,item)
        
        return handlerInput.responseBuilder.speak(speak).reprompt(responses["WELCOME_REPROMPT"]!).getResponse()
    }
}

struct HelpHandler: RequestHandler {
    
    func canHandle(handlerInput: HandlerInput) -> Bool {
        return (handlerInput.requestEnvelope.request.type == "IntentRequest" && handlerInput.requestEnvelope.request.intent!.name == "AMAZON.HelpIntent")
    }
    
    func handle(handlerInput: HandlerInput) -> Response {
        
        let item:String = recipes.randomElement()!.key
        
        let speak = String(format: responses["HELP_MESSAGE"]!,item)
        
        return handlerInput.responseBuilder.speak(speak).reprompt(responses["HELP_REPROMPT"]!).getResponse()
    }
}

struct RepeatHandler: RequestHandler {
    
    func canHandle(handlerInput: HandlerInput) -> Bool {
        return (handlerInput.requestEnvelope.request.type == "IntentRequest" && handlerInput.requestEnvelope.request.intent!.name == "AMAZON.RepeatIntent")
    }
    
    func handle(handlerInput: HandlerInput) -> Response {
        
        let speakOutput: String = handlerInput.attributeMananger.getSessionAttributes()["speakOutput"] as! String
        let reprompt: String? = handlerInput.attributeMananger.getSessionAttributes()["repromptSpeech"] as? String
        
        let builder = handlerInput.responseBuilder.speak(speakOutput)
            
        if (reprompt != nil) {
            builder.reprompt(reprompt!)
        }
            
        return builder.getResponse()
    }
}

struct ExitHandler: RequestHandler {
    
    func canHandle(handlerInput: HandlerInput) -> Bool {
        if (handlerInput.requestEnvelope.request.type == "IntentRequest") {
            return (handlerInput.requestEnvelope.request.intent!.name == "AMAZON.StopIntent") || (handlerInput.requestEnvelope.request.intent!.name == "AMAZON.CancelIntent")
        } else {
            return false
        }
    }
    
    func handle(handlerInput: HandlerInput) -> Response {
        
        let speak = String(format: responses["STOP_MESSAGE"]!)
        
        return handlerInput.responseBuilder.speak(speak).getResponse()
    }
}

struct ErrorHandler: RequestHandler {
    
    func canHandle(handlerInput: HandlerInput) -> Bool {
        return true
    }
    
    func handle(handlerInput: HandlerInput) -> Response {
        handlerInput.responseBuilder.speak("Sorry, I can\'t understand the command. Please say again.").reprompt("Sorry, I can\'t understand the command. Please say again.").getResponse()
    }
}

enum DialogError: Error {
    case unhandled
}


class SkillDialogManager {
    var session:[String:Any] = [:]
    
    var handlers:[RequestHandler] = [LaunchRequestHandler(),HelpHandler(),RepeatHandler(),ExitHandler(),ErrorHandler()]
    
    func turn(type:String, nluResult:NLUResult? = nil, error:Error? = nil) throws -> Response {
        
        let handlerInput = HandlerInput(session: session, type: type, nluResult: nluResult)
        
        for handler in handlers {
            if (handler.canHandle(handlerInput: handlerInput)) {
                let output = handler.handle(handlerInput: handlerInput)
                session["speakOutput"] = output.speak
                session["repromptSpeech"] = output.reprompt
                
                return output
            }
        }
        
        throw DialogError.unhandled
    }
}
